.setcpu "6502"
.autoimport on

;iNESヘッダ
.segment "HEADER"
	.byte	$4e, $45, $53, $1a	;"NES"
	.byte	$02	;プログラムバンク数
	.byte	$01	;キャラクタバンク数
	.byte	$01	;水平ミラー
	.byte	$00
	.byte	$00, $00, $00, $00
	.byte	$00, $00, $00, $00

.segment "STARTUP"

;リセット割り込み
.proc	Reset
	sei
	ldx	#$ff
	txs

;スクリーンオフ
	lda	#$00
	sta	$2000
	sta	$2001

;ゼロページとネームテーブルの上の部分を0で初期化
	lda #$20
	sta $2006
	lda #$00
	sta $2006
	ldx #$00
	ldy #$00
ResetZeroPage:
	sta $00,x
ResetNameTable:
	sta $2007
	iny
	bne ResetNameTable
	inx
	bne ResetZeroPage


;パレットテーブルへ転送
	lda	#$3f
	sta	$2006
	lda	#$00
	sta	$2006
	ldx #$00
	ldy #$20
copypal:
	lda palettes, x
	sta $2007
	inx
	dey
	bne copypal

;ネームテーブルへ転送
	lda #$20
	sta $2006
	lda #$22
	sta $2006
	ldx #$00
	ldy #$14
;スコア欄の画像を転送する
;スコアはただの飾り
copymap1:
	lda scoretable,x
	sta $2007
	inx
	dey
	bne copymap1

;上付きバーの表示
	lda #$20
	sta $2006
	lda #$40
	sta $2006
	lda #$12
	ldx #$20
copymap2:
	sta $2007
	dex
	bne copymap2

;属性テーブルへ転送
	lda #$23
	sta $2006
	lda #$c0
	sta $2006
	ldy #$00
	lda ROW
;二重ループ
copyAttr:
	pha
	lda blockcolor, y
	ldx COLUMN
copyColumn:
	sta $2007
	dex
	bne copyColumn
	pla
	iny
	sec
	sbc #$01
	bne copyAttr

;スクロール設定
	lda #$00
	sta $2005
	sta $2005

;スクリーンオン
	lda #$88
	sta $2000
	lda #$1e
	sta $2001

;無限ループ
;VBlank割り込みでメインループに飛ぶ
infinityloop:
	jmp infinityloop
.endproc

;変数定義
GameState	=	$00
Bar_Pos_X	=	$01
Bar_Speed	=	$02
Ball_Pos_X	=	$03
Ball_Pos_Y	=	$04
Ball_Vel_X	=	$05
Ball_Vel_Y	=	$06
Ball_State	=	$07
Ball_Lives	=	$08
BlockRefPos	=	$09	;ブロックの描画位置の指定&ブロックの衝突判定位置
BlockRemain	=	$0a
Key_A		=	$0b
Key_Start	=	$0c
Key_Left	=	$0d
Key_Right	=	$0e
BlockState	=	$10	;ここから$80Byte分($10~$8f)がブロック
MultiX		=	$90
MultiY		=	$91
Dividend	=	$92
DivMod		=	$93
DivQuo		=	$94
DivRem		=	$95

;定数定義
;バーのy座標(位置は#$cf)
Bar_Pos_Y:	.byte $cf
;属性テーブル行数
ROW:		.byte $03
;属性テーブル列数
COLUMN:		.byte $08
;ブロックの行数
BLOCK_ROW:	.byte $80
;分岐用定数
INIT_GAME:	.byte $00
GAME_MAIN:	.byte $01
GAME_OVER:	.byte $02
GAME_CLEAR:	.byte $03
;ボールの状態
BALL_ON_THE_PLATE:	.byte $00
BALL_CASTED:		.byte $01
BALL_FALL:			.byte $02

;メインループ
.proc	mainloop
	;キー入力
	lda #$01
	sta $4016
	lda #$00
	sta $4016
	lda $4016
	and #$01
	sta Key_A
	lda $4016
	lda $4016
	lda $4016
	and #$01
	sta Key_Start
	lda $4016
	lda $4016
	lda $4016
	and #$01
	sta Key_Left
	lda $4016
	and #$01
	sta Key_Right
	;フラグによる分岐（初期化、ゲームメイン、ゲームオーバー、ゲームクリア）
	;初期化
	lda GameState
	cmp INIT_GAME
	bne SkipInitGame
	jsr	InitGame
	jmp SkipGameClear
SkipInitGame:

	;ゲームメイン
	cmp GAME_MAIN
	bne SkipGameMain
	jsr	GameMain
	jmp SkipGameClear
SkipGameMain:

	;ゲームオーバー
	cmp GAME_OVER
	bne SkipGameOver
	jsr	GameOver
	jmp SkipGameClear
SkipGameOver:

	;ゲームクリア
	cmp GAME_CLEAR
	bne SkipGameClear
	jsr	GameClear
SkipGameClear:

	;スクロール位置の設定
	lda #$00
	sta $2005
	sta $2005
	rti
.endproc

;ブロック描画処理
DrawBlock:
	lda #$20
	sta $2006
	lda BlockRefPos
	asl
	clc
	adc #$80
	sta $2006
	;オーバーフローしたら
	bvc noCarry
	lda #$21
	sta $2006
	lda BlockRefPos
	asl
	clc
	adc #$80
	sta $2006
noCarry:
	;ブロックの描画
	ldx BlockRefPos
	lda BlockState, x
	and #$01
	beq SkipDrawBlock
	lda #$02
	sta $2007
	lda #$03
	sta $2007
	cmp #$01;ゼロフラグをクリアする
SkipDrawBlock:
	bne SkipEraceBlock
	lda #$00
	sta $2007
	sta $2007
SkipEraceBlock:
	rts

;ゲームの初期化処理
InitGame:
	;変数の初期化
	lda #$82
	sta Bar_Pos_X
	lda #$02
	sta Bar_Speed
	lda #$03
	sta Ball_Lives
	lda #$80
	sta BlockRemain
	lda #$01
	ldx BlockRefPos
	sta BlockState,x
	;スプライトDMA用の設定
	;ボール
	lda #$02
	sta $7001
	lda #%00000000
	sta $7002
	;バー
	;一つ目
	lda #$00
	sta $7005
	lda #%00000000
	sta $7006
	;二つ目
	lda #$01
	sta $7009
	lda #%00000000
	sta $700a
	;三つ目
	lda #$01
	sta $700d
	lda #%00000000
	sta $700e
	;四つ目
	lda #$00
	sta $7011
	lda #%01000000
	sta $7012
	;ブロックをずらーって描画
	jsr DrawBlock
	inc BlockRefPos

	;描画終わったらゲームメインへ
	lda BlockRefPos
	cmp BLOCK_ROW
	bne SkipSwitchGameState
	lda GAME_MAIN
	sta GameState
SkipSwitchGameState:
	rts

;ゲームオーバー処理
GameOver:
	ldx #$00
	ldy #$0b
	lda #$21
	sta $2006
	lda #$aa
	sta $2006
DrawGameOver:
	lda String_GameOver,x
	sta $2007
	inx
	dey
	bne DrawGameOver
	rts

GameClear:
	ldx #$00
	ldy #$0b
	lda #$21
	sta $2006
	lda #$aa
	sta $2006
DrawGameClear:
	lda String_GameClear,x
	sta $2007
	inx
	dey
	bne DrawGameClear
	rts

;ゲームのメイン部分
GameMain:
	lda Ball_State
	cmp BALL_ON_THE_PLATE
	bne SkipBallOnThePlate
	;ボールの状態がBALL_ON_THE_PLATEなら板の上にボールをのせる
	jsr GameMain_OnThePlate
	jmp SkipBallFall
SkipBallOnThePlate:

	cmp BALL_CASTED
	bne SkipBallCasted
	;ボールの状態が01なら
	jsr GameMain_Casted
	jmp SkipBallFall
SkipBallCasted:

	cmp BALL_FALL
	bne SkipBallFall
	;ボールの状態が02なら
	jsr GameMain_Fall
SkipBallFall:

	;ボールの描画
	lda Ball_Pos_Y
	sta $7000
	lda  Ball_Pos_X
	sta $7003
	;バーの描画
	;一つ目
	lda Bar_Pos_Y
	sta $7004
	lda  Bar_Pos_X
	sta $7007
	;二つ目
	lda Bar_Pos_Y
	sta $7008
	lda  Bar_Pos_X
	clc
	adc #$08
	sta $700b
	;三つ目
	lda Bar_Pos_Y
	sta $700c
	lda  Bar_Pos_X
	clc
	adc #$10
	sta $700f
	;四つ目
	lda Bar_Pos_Y
	sta $7010
	lda  Bar_Pos_X
	clc
	adc #$18
	sta $7013
	;スプライトDMA転送
	lda #$70
	sta $4014

	rts

;ボールがバーの上に乗っている状態
GameMain_OnThePlate:
	;バーの移動
	lda Key_Left
	cmp #$01
	bne SkipKeyLeft1
	lda Bar_Pos_X
	sec
	sbc Bar_Speed
	sta Bar_Pos_X
SkipKeyLeft1:

	lda Key_Right
	cmp #$01
	bne SkipKeyRight1
	lda Bar_Pos_X
	clc
	adc Bar_Speed
	sta Bar_Pos_X
SkipKeyRight1:

	;ボールの速度&位置指定
	lda #$02
	sta Ball_Vel_X
	lda #$02
	jsr invert
	sta Ball_Vel_Y
	lda #$c8
	sta Ball_Pos_Y
	;ボールをバーに追従させる
	lda Bar_Pos_X
	clc
	adc #$0c
	sta Ball_Pos_X

	;キーが押されたら玉を発射
	lda Key_A
	cmp #$01
	bne SkipKeyA
	lda BALL_CASTED
	sta Ball_State
SkipKeyA:
	rts
GameMain_Casted:
	;バーの移動
	lda Key_Left
	cmp #$01
	bne SkipKeyLeft2
	lda Bar_Pos_X
	sec
	sbc Bar_Speed
	sta Bar_Pos_X
SkipKeyLeft2:

	lda Key_Right
	cmp #$01
	bne SkipKeyRight2
	lda Bar_Pos_X
	clc
	adc Bar_Speed
	sta Bar_Pos_X
SkipKeyRight2:

	;ボールの移動
	lda Ball_Pos_X
	clc
	adc Ball_Vel_X
	sta Ball_Pos_X

	lda Ball_Pos_Y
	clc
	adc Ball_Vel_Y
	sta Ball_Pos_Y

	;板との当たり判定
	;板とのy座標差が-5ドット以上5ドット以下
	lda Ball_Pos_Y
	sec
	sbc Bar_Pos_Y
	;絶対値
	jsr abs
	sec
	sbc #$05
	bcs SkipBarHit
	;x座標差が-16ドット以上16ドット以下
	;|ball_x+4-(bar_x+16)|より|ball_x-bar_x-12|
	lda Ball_Pos_X
	sec
	sbc Bar_Pos_X
	sec
	sbc #$0c
	;絶対値
	jsr abs
	sec
	sbc #$10
	bcs SkipBarHit
	;バーとボールのx座標の差が
	lda Ball_Pos_X
	sec
	sbc Bar_Pos_X
	clc
	adc #$04
	;8以下なら左に2の速さ
	cmp #$8
	bcs Above8
		lda #$02
		jsr invert
		sta Ball_Vel_X
		jmp InvertVelocity_Y
	Above8:
	;8以上16以下なら左に1の速さ
	cmp #$10
	bcs Above16
		lda #$01
		jsr invert
		sta Ball_Vel_X
		jmp InvertVelocity_Y
	Above16:
	;16以上24以下なら右に1の速さ
	cmp #$18
	bcs Above24
		lda #$01
		sta Ball_Vel_X
		jmp InvertVelocity_Y
	Above24:
	;24以上なら右に2の速さ
	lda #$02
	sta Ball_Vel_X
	InvertVelocity_Y:
	;速度のy成分を負に
	lda #$02
	jsr invert
	sta Ball_Vel_Y
SkipBarHit:

	;ブロックとの当たり判定
	;ボールのいる位置にブロックがあるなら
	lda Ball_Pos_Y
	clc
	adc #$07		;ボールの中心に移すための4ドット＋謎の3ドット
	sec
	sbc #$20		;ボールの全ブロックの左上の座標からの相対位置;上8ドットは隠れているゆえその分も考える
	bcc SkipHitBlock
	cmp #$40
	bcs SkipHitBlock
	sta Dividend
	lda #$8
	sta DivMod
	jsr div
	lda DivQuo
	sta MultiX
	lda #$10
	sta MultiY
	jsr mul
	;ここまででy座標のブロックの位置が出る
	tax
	lda Ball_Pos_X
	clc
	adc #$04		;ボールの中心に移すための4ドット
	sta Dividend
	lda #$10
	sta DivMod
	jsr div
	txa
	clc
	adc DivQuo
	;ブロックがあったらブロックを消す
	sta BlockRefPos
	tax
	lda BlockState,x
	and #$01
	beq SkipHitBlock
	dec BlockRemain
	lda #$00
	sta BlockState,x
	jsr DrawBlock
	;跳ね返り処理
	jsr Hitting_Block
	;ブロック数が0ならGameStateをGAME_CLEARに
	lda BlockRemain
	bne SkipHitBlock
	lda GAME_CLEAR
	sta GameState
SkipHitBlock:
	;壁との当たり判定
	;x座標が0または252なら速度x成分反転
	lda Ball_Pos_X
	sec
	sbc #$02
	bcs SkipTurnVelocX1
	lda Ball_Vel_X
	jsr invert
	sta Ball_Vel_X
SkipTurnVelocX1:

	lda Ball_Pos_X
	clc
	adc #$08
	bcc SkipTurnVelocX2
	lda Ball_Vel_X
	jsr invert
	sta Ball_Vel_X
SkipTurnVelocX2:
	;y座標が8ならy成分反転
	lda Ball_Pos_Y
	sec
	sbc #$10
	bcs SkipTurnVelocY
	lda Ball_Vel_Y
	jsr invert
	sta Ball_Vel_Y
SkipTurnVelocY:
	;y座標が254ならBall_StateをBALL_FALLに
	lda Ball_Pos_Y
	clc
	adc #$02
	bcc SkipGoFall
	lda BALL_FALL
	sta Ball_State
SkipGoFall:
	rts

;ブロックと衝突した際の跳ね返りの判定
Hitting_Block:
	;調べるブロックを指定（サブルーチン元で）
	lda BlockRefPos
	;ブロックの中心から見たボールの中心の相対的な距離を取得
	;ブロックのx座標を得る
	and #$0f
	sta MultiX
	lda #$10
	sta MultiY
	jsr mul
	;ボールとの相対距離x取得
	;|block_x+8-(ball_x+4)|より|block_x+4-ball_x|
	clc
	adc #$04
	sec
	sbc Ball_Pos_X
	jsr abs
	tax		;x座標をxレジスタに退避しておく
	;ブロックのy座標を得る
	lda BlockRefPos
	and #$f0
	sta Dividend
	lda #$2
	sta DivMod
	jsr div
	lda DivQuo
	clc
	adc #$20
	;ボールとの相対距離y取得
	;|block_y+4-(ball_y+4+(謎の3ドット))|より|block_y-3-ball_y|
	sec
	sbc #$03
	sec
	sbc Ball_Pos_Y
	php
	jsr abs
	plp
	beq reflectX
	;y相対距離を2倍に
	sta MultiX
	lda #$02
	sta MultiY
	jsr mul
	;x/yを計算
	sta DivMod
	txa
	sta Dividend
	jsr div
	lda DivQuo
	;1引いてキャリーフラグが0だったら(ボロ―が起きたら)y反転
	sec
	sbc #$01
	bcc reflectY
	;ゼロフラグが立っていなかったらx反転
	bne reflectX
	;ゼロフラグが立っていてかつ余りが0でないならx反転
	lda DivRem
	bne reflectX
	;余りが0ならxy反転
	;両方に反射
reflectXY:
	;余り無し1なら両方反射
	lda Ball_Vel_X
	jsr invert
	sta Ball_Vel_X
	lda Ball_Vel_Y
	jsr invert
	sta Ball_Vel_Y
	rts
	;横に反射
reflectX:
	;yが0なら横に反射
	;x/yが1より大きいなら横に反射
	lda Ball_Vel_X
	jsr invert
	sta Ball_Vel_X
	rts
	;縦に反射
reflectY:
	;1より小さいなら縦に反射
	lda Ball_Vel_Y
	jsr invert
	sta Ball_Vel_Y
	rts

;ボールが落ちたときの処理
GameMain_Fall:
	;残機0ならゲームオーバー
	lda Ball_Lives
	bne SkipGameOver
	lda GAME_OVER
	sta GameState
	rts
SkipGameOver:
	;ボールの状態をBALL_ON_THE_PLATEに
	dec Ball_Lives
	lda #$20
	sta $2006
	lda #$23
	clc
	adc Ball_Lives
	sta $2006
	lda #$00
	sta $2007

	lda BALL_ON_THE_PLATE
	sta Ball_State
	rts

;aの正負を反転
invert:
	eor #$ff
	clc
	adc #$01
	rts

;絶対値
;絶対値を取りたい減算の後にjsr abs
abs:
	bpl SkipAbs
	jsr invert
SkipAbs:
	rts

;掛け算
;MultiXとMultiYに値を入れて掛け算してもらう
mul:
	lda #$00
	mulloop:
	clc
	adc MultiX
	dec MultiY
	bne mulloop
	rts

;割り算
;Dividendに被除数、DivModに除数を入れる
;DivQuoに商、DivRemに余りを入れて返す
div:
	lda #$00
	sta DivQuo
	lda Dividend
	divloop:
	inc DivQuo
	sec
	sbc DivMod
	bcs divloop
	adc DivMod
	sta DivRem
	dec DivQuo
	rts


;パレットテーブル
palettes:
	.byte	$0f, $08, $18, $29
	.byte	$0f, $0a, $1a, $2c
	.byte	$0f, $17, $28, $39
	.byte	$0f, $06, $16, $26

	.byte	$0c, $17, $28, $39
	.byte	$0f, $06, $16, $26
	.byte	$0f, $08, $18, $28
	.byte	$0f, $0a, $1a, $2a

;スコアテーブル
scoretable:
	.byte	$00, $01, $01, $01
	.byte	$00, $00, $0f, $08
	.byte	$0d, $0e, $09, $11
	.byte	$04, $04, $04, $04
	.byte	$04, $04, $04, $04

;ブロックカラーテーブル
blockcolor:
	.byte	$aa, $af, $05

;GAMEOVERの文字
String_GameOver:
	.byte	$0a, $07, $0c, $09
	.byte	$00, $00, $0d, $10
	.byte	$09, $0e, $06

;GAMECLEARの文字
String_GameClear:
	.byte	$0a, $07, $0c, $09
	.byte	$00, $08, $0b, $09
	.byte	$07, $0e, $06

.segment "VECINFO"
	.word mainloop
	.word Reset
	.word $0000

;パターンテーブル
.segment "CHARS"
	.incbin "Arkanoid.chr"
